{

  Chat codes v1.2.1
  ================
  New in v1.2
    - \s command (font size)
    - preview window
    - emoticons
    - url detection
  New in v1.1
    - \f command (font name)
}

unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, RVScroll, RichView, RVStyle, RVFuncs, RVItem,
  CRVData, CRVFData, RVFileFuncs, RVGrHandler, RVTypes, RVStrFuncs, 
  ShellApi;

type
  TEmoticonCode = record
    Code: TRVUnicodeString;
    Index: Integer;
  end;

  TForm1 = class(TForm)
    rvs: TRVStyle;
    rv: TRichView;
    Edit: TEdit;
    btnEnter: TButton;
    btnBold: TButton;
    btnItalic: TButton;
    btnUnderline: TButton;
    btnColor: TButton;
    btnBkColor: TButton;
    ColorDialog1: TColorDialog;
    Label1: TLabel;
    btnArial: TButton;
    btnTimes: TButton;
    btnFontSize: TButton;
    rvPreview: TRichView;
    Label2: TLabel;
    procedure btnBoldClick(Sender: TObject);
    procedure btnItalicClick(Sender: TObject);
    procedure btnUnderlineClick(Sender: TObject);
    procedure btnColorClick(Sender: TObject);
    procedure btnBkColorClick(Sender: TObject);
    procedure btnEnterClick(Sender: TObject);
    procedure EditKeyPress(Sender: TObject; var Key: Char);
    procedure btnArialClick(Sender: TObject);
    procedure btnTimesClick(Sender: TObject);
    procedure btnFontSizeClick(Sender: TObject);
    procedure EditChange(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure rvJump(Sender: TObject; id: Integer);
  private
    { Private declarations }
    EmoticonsGraphics: array of TGraphic;
    EmoticonsCodes: array of TEmoticonCode;
    function GetStyleNo(CommandChar: Char; Color: TColor; StyleNo: Integer;
      const FontName: String; FontSize: Integer): Integer;
    function GetHypertextStyleNo(StyleNo: Integer): Integer;
    procedure ParseString(rv: TCustomRichView; s: TRVUnicodeString);
    procedure CreateEmoticons;
    procedure DestroyEmoticons;
    function EmoticonAtIndex(const s: TRVUnicodeString; Index: Integer): Integer;
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{============================ EMOTICONS =======================================}
procedure TForm1.FormCreate(Sender: TObject);
begin
  CreateEmoticons;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  DestroyEmoticons;
end;
{------------------------------------------------------------------------------}
// Initializing emoticons arrays
procedure TForm1.CreateEmoticons;
var i: Integer;
begin
  // EmoticonsGraphics array contains graphics for emoticons.
  // This code loads 0.bmp, 1.bmp, ..., 3.bmp as TBitmaps from Emoticons subfolder
  SetLength(EmoticonsGraphics, 4);
  for i := 0 to High(EmoticonsGraphics) do
  begin
    EmoticonsGraphics[i] := TBitmap.Create;
    EmoticonsGraphics[i].LoadFromFile(ExtractFilePath(Application.ExeName)+
      'Emoticons\'+IntToStr(i)+'.bmp');
  end;
  // Items in EmoticonsCodes array have the properties:
  // - Code - emoticon text
  // - Index - index in EmoticonsGraphics array
  SetLength(EmoticonsCodes, 8);
  EmoticonsCodes[0].Code := ':)';
  EmoticonsCodes[0].Index := 0;
  EmoticonsCodes[1].Code := ':-)';
  EmoticonsCodes[1].Index := 0;
  EmoticonsCodes[2].Code := ':(';
  EmoticonsCodes[2].Index := 1;
  EmoticonsCodes[3].Code := ':-(';
  EmoticonsCodes[3].Index := 1;
  EmoticonsCodes[4].Code := ':D';
  EmoticonsCodes[4].Index := 2;
  EmoticonsCodes[5].Code := ':-D';
  EmoticonsCodes[5].Index := 2;
  EmoticonsCodes[6].Code := ';)';
  EmoticonsCodes[6].Index := 3;
  EmoticonsCodes[7].Code := ';-)';
  EmoticonsCodes[7].Index := 3;
end;
{------------------------------------------------------------------------------}
// Destroying emoticons arrays
procedure TForm1.DestroyEmoticons;
var i: Integer;
begin
  for i := 0 to High(EmoticonsGraphics) do
    EmoticonsGraphics[i].Free;
  SetLength(EmoticonsGraphics, 0);
  SetLength(EmoticonsCodes, 0);
end;
{------------------------------------------------------------------------------}
// Returns index in EmoticonsCodes array, if s contains emoticon code starting
// from the character Index. Otherwise, returns -1.
function TForm1.EmoticonAtIndex(const s: TRVUnicodeString; Index: Integer): Integer;
var i: Integer;
begin
  for i := 0 to High(EmoticonsCodes) do
    if Copy(s, Index, Length(EmoticonsCodes[i].Code))=EmoticonsCodes[i].Code then
    begin
      Result := i;
      exit;
    end;
  Result := -1;
end;

{$R *.dfm}
{==================== INSERTING CHAT CODES ====================================}
// TColor to RGB string
function GetColorStr(Color: TColor): TRVUnicodeString;
begin
  Result := IntToHex(ColorToRGB(Color),6);
  Result := Copy(Result,5,2)+Copy(Result,3,2)+Copy(Result,1,2);
end;
// RGB string to TColor
function GetStrColor(ColorValue: TRVUnicodeString): TColor;
var i,v: Integer;
begin
  ColorValue := Copy(ColorValue,5,2)+Copy(ColorValue,3,2)+Copy(ColorValue,1,2);
  Result := 0;
  for i := 1 to 6 do
  begin
    if RVCharInSet7(ColorValue[i], ['A'..'F']) then
      v := ord(ColorValue[i])-ord('A')+10
    else if RVCharInSet7(ColorValue[i], ['a'..'f']) then
      v := ord(ColorValue[i])-ord('a')+10
    else
      v := ord(ColorValue[i])-ord('0');
    Result := Result*16+v;
  end;
end;
{------------------------------------------------------------------------------}
// Insert command: bold on/off
procedure TForm1.btnBoldClick(Sender: TObject);
begin
  Edit.SelText := '\b';
  Edit.SetFocus;
end;
{------------------------------------------------------------------------------}
// Insert command: italic on/off
procedure TForm1.btnItalicClick(Sender: TObject);
begin
  Edit.SelText := '\i';
  Edit.SetFocus;
end;
{------------------------------------------------------------------------------}
// Insert command: underline on/off
procedure TForm1.btnUnderlineClick(Sender: TObject);
begin
  Edit.SelText := '\u';
  Edit.SetFocus;
end;
{------------------------------------------------------------------------------}
// Insert command: text color
procedure TForm1.btnColorClick(Sender: TObject);
begin
  ColorDialog1.Color := clBlack;
  if ColorDialog1.Execute then
    Edit.SelText := '\c'+GetColorStr(ColorDialog1.Color);
  Edit.SetFocus;
end;
{------------------------------------------------------------------------------}
// Insert command: text background color
procedure TForm1.btnBkColorClick(Sender: TObject);
begin
  ColorDialog1.Color := clWhite;
  if ColorDialog1.Execute then
    Edit.SelText := '\k'+GetColorStr(ColorDialog1.Color);
  Edit.SetFocus;
end;
{------------------------------------------------------------------------------}
// Insert command: font Arial
procedure TForm1.btnArialClick(Sender: TObject);
begin
  Edit.SelText := '\f"Arial"';
  Edit.SetFocus;
end;
{------------------------------------------------------------------------------}
// Insert command: font Times
procedure TForm1.btnTimesClick(Sender: TObject);
begin
  Edit.SelText := '\f"Times New Roman"';
  Edit.SetFocus;
end;
{------------------------------------------------------------------------------}
// Insert command: font size
procedure TForm1.btnFontSizeClick(Sender: TObject);
var s: String;
begin
  s := '8';
  if not InputQuery('Enter Font Size', '&Size:', s) then
    exit;
  if Length(s)=1 then
    s := '0'+s;
  if Length(s)>2 then
    s := Copy(s, 1, 2);
  Edit.SelText := '\s'+s;
  Edit.SetFocus;
end;
{==================== PROCESSING EDITOR EVENTS ================================}
// Adding string in RichView
procedure TForm1.btnEnterClick(Sender: TObject);
begin
  ParseString(rv, Edit.Text);
  rv.FormatTail;
  rv.Update;
  Edit.Text := '';
  rv.DeleteUnusedStyles(True, True, True);
  Edit.SetFocus;
end;
{------------------------------------------------------------------------------}
// Updating preview window
procedure TForm1.EditChange(Sender: TObject);
begin
  rvPreview.Clear;
  ParseString(rvPreview, Edit.Text);
  rvPreview.Format;
  rvPreview.Update;
end;
{------------------------------------------------------------------------------}
// Adding string when pressing Enter in the editor
procedure TForm1.EditKeyPress(Sender: TObject; var Key: Char);
begin
  if Key=#13 then
  begin
    Key := #0;
    btnEnterClick(Sender);
  end;
end;
{------------------------------------------------------------------------------}
// Returning StyleNo basing on the command:
// 'b' - bold on/off;          - Color, FontName and FontSize are ignored
// 'i' - italic on/off;        - Color, FontName and FontSize are ignored
// 'u' - underline on/off;     - Color, FontName and FontSize are ignored
// 'c' - set text Color;       - FontName and FontSize are ignored
// 'k' - set background Color; - FontName and FontSize are ignored
// 'f' - font name;            - Color and FontSize are ignored
// 's' - font size;            - Color and FontName are ignored
function TForm1.GetStyleNo(CommandChar: Char; Color: TColor;
  StyleNo: Integer; const FontName: String; FontSize: Integer): Integer;
var fi: TFontInfo;
begin
  fi := TFontInfo.Create(nil);
  fi.Assign(rvs.TextStyles[StyleNo]);
  case CommandChar of
    'b':
      if fsBold in fi.Style then
        fi.Style := fi.Style-[fsBold]
      else
        fi.Style := fi.Style+[fsBold];
    'i':
      if fsItalic in fi.Style then
        fi.Style := fi.Style-[fsItalic]
      else
        fi.Style := fi.Style+[fsItalic];
    'u':
      if fsUnderline in fi.Style then
        fi.Style := fi.Style-[fsUnderline]
      else
        fi.Style := fi.Style+[fsUnderline];
    'c':
      fi.Color := Color;
    'k':
      fi.BackColor := Color;
    'f':
      fi.FontName := FontName;
    's':
      fi.Size := FontSize;
  end;
  Result := rvs.FindTextStyle(fi);
  fi.Free;
end;
{------------------------------------------------------------------------------}
// Returns index of text styles having all properties of the StyleNo-th style,
// but hypertext, blue, highlighted and underlined
function TForm1.GetHypertextStyleNo(StyleNo: Integer): Integer;
var fi: TFontInfo;
begin
  fi := TFontInfo.Create(nil);
  fi.Assign(rvs.TextStyles[StyleNo]);
  fi.Jump := True;
  fi.Color := clBlue;
  fi.HoverColor := clWhite;
  fi.HoverBackColor := clBlue;
  fi.Style := fi.Style+[fsUnderline];
  Result := rvs.FindTextStyle(fi);
  fi.Free;
end;
{------------------------------------------------------------------------------}
// Returns position of the first delimiter character in pc, from 1 (or 0 is none)
function FindChar(pc: PRVUnicodeChar; Len: Integer; rv: TCustomRichView): Integer;
var i: Integer;
begin
  for i := 0 to Len-1 do
    if not RVCharInSet7(pc[i], ['.', ':', '/']) and rv.RVData.IsDelimiterW(pc[i]) then
    begin
      Result := i+1;
      exit;
    end;
  Result := 0;
end;
{------------------------------------------------------------------------------}
type TReadState = (rsNormal, rsFontNeedsFirstQuote, rsFontName, rsColor, rsFontSize);
// Adding string with codes
procedure TForm1.ParseString(rv: TCustomRichView; s: TRVUnicodeString);
var i: Integer;
    ParaNo, StyleNo, StartIndex, EmoticonIndex: Integer;
    CodeStarted: Boolean;
    ReadState: TReadState;
    ColorChar: Char;
    Value: TRVUnicodeString;
    {..................................}
    // replacing '\\' with '\'
    procedure RemoveDoubleBS(var str: TRVUnicodeString);
    var i: Integer;
        BSDeleted: Boolean;
    begin
      BSDeleted := False;
      for i := Length(str) downto 1 do
        if (str[i]='\') and (i>1) and (str[i-1]='\') and not BSDeleted then
        begin
          Delete(str, i, 1);
          BSDeleted := True;
        end
        else
          BSDeleted := False;
    end;
    {..................................}
    // Adds text from s[StartIndex] to s[EndIndex] in rv, with URL detection
    // Uses StyleNo and ParaNo. Changes ParaNo.
    procedure AddStringWithURLs(StartIndex, EndIndex: Integer);
    var str, CurrentWord, s1: TRVUnicodeString;
        p: Integer;
        pc, pcstart: PRVUnicodeChar;
        Len: Integer;
        HypertextStyleNo: Integer;
    begin
      if EndIndex<StartIndex then
        exit;
      HypertextStyleNo := -1;
      str := Copy(s, StartIndex, EndIndex-StartIndex+1);
      RemoveDoubleBS(str);
      pc := PRVUnicodeChar(str);
      pcstart := pc;
      Len := Length(str);
      while Len>0 do
      begin
        p := FindChar(pc, Len, rv);
        if p=1 then
        begin
          inc(pc);
          dec(Len);
          continue;
        end;
        if p=0 then
          p := Len+1;
        SetLength(CurrentWord, p-1);
        Move(pc^, PRVUnicodeChar(CurrentWord)^, (p-1)*sizeof(TRVUnicodeChar));
        if (RVIsURL(CurrentWord) or RVIsEmail(CurrentWord)) then
        begin
          if pcstart<pc then
          begin
            SetLength(s1, pc-pcstart);
            Move(pcstart^, PRVUnicodeChar(s1)^, (pc-pcstart)*sizeof(TRVUnicodeChar));
            rv.AddNLW(s1, StyleNo, ParaNo);
            ParaNo := -1;
          end;
          if HypertextStyleNo<0 then
            HypertextStyleNo := GetHypertextStyleNo(StyleNo);
          rv.AddNLW(CurrentWord, HypertextStyleNo, ParaNo);
          ParaNo := -1;
          inc(pc, p-1);
          dec(Len, p-1);
          pcstart := pc;
        end
        else
        begin
          inc(pc, p-1);
          dec(Len, p-1);
        end;
      end;
      if pcstart<pc then
      begin
        SetLength(s1, pc-pcstart);
        Move(pcstart^, PRVUnicodeChar(s1)^, (pc-pcstart)*sizeof(TRVUnicodeChar));
        rv.AddNLW(s1, StyleNo, ParaNo);
        ParaNo := -1;
      end;
    end;
    {..................................}
    // Adds the emoticon defined in EmoticonsCodes[Index]
    procedure AddEmoticon(Index: Integer);
    var gr: TGraphic;
    begin
      gr := RVGraphicHandler.CreateGraphic(
        TGraphicClass(EmoticonsGraphics[EmoticonsCodes[Index].Index].ClassType));
      gr.Assign(EmoticonsGraphics[EmoticonsCodes[Index].Index]);
      rv.AddPicture( '', gr, ParaNo, rvvaBaseline);
      rv.SetItemExtraStrProperty(rv.ItemCount-1, rvespAlt, EmoticonsCodes[Index].Code);
      ParaNo := -1;
    end;
    {..................................}
begin
  ParaNo := 0;
  CodeStarted := False;
  StyleNo := 0;
  StartIndex := 1;
  ColorChar := #0;
  ReadState := rsNormal;
  for i := 1 to Length(s) do
    if i>=StartIndex then
      case ReadState of
        rsColor: // expecting color value, RRGGBB
          begin
            if RVCharInSet7(s[i], ['0'..'9', 'a'..'f', 'A'..'F']) then
            begin
              Value := Value + s[i];
              if Length(Value)=6 then
              begin
                AddStringWithURLs(StartIndex, i-2-6);
                StartIndex := i+1;
                StyleNo := GetStyleNo(ColorChar, GetStrColor(Value), StyleNo, '', 0);
                ColorChar := #0;
                ReadState := rsNormal;
              end;
              continue;
            end;
            ReadState := rsNormal;
          end;
        rsFontNeedsFirstQuote: // expecting opening quote for the font name
          begin
            if s[i]<>'"' then
              ReadState := rsNormal
            else
              ReadState := rsFontName;
          end;
        rsFontName: // expecting font name, till the closing quote
          begin
            if s[i]<>'"' then
              Value := Value+s[i]
            else
            begin
              AddStringWithURLs(StartIndex, i-2-2-Length(Value));
              ReadState := rsNormal;
              StyleNo := GetStyleNo('f', clNone, StyleNo, Value, 0);
              StartIndex := i+1;
            end;
          end;
        rsFontSize: // expecting font size (either 2 digits, or 1 digit then nondigit;
                    // if the latter, then if this nondigit is a space-char,
                    // it will not be added in text)
          begin
            if RVCharInSet7(s[i], ['0'..'9']) then
              Value := Value + s[i];
            if (Length(Value)=2) or (i=Length(s)) or not RVCharInSet7(s[i], ['0'..'9']) then
            begin
              ReadState := rsNormal;
              if Length(Value)>0 then
              begin
                if RVCharInSet7(s[i], ['0'..'9']) then
                  AddStringWithURLs(StartIndex, i-2-Length(Value))
                else
                  AddStringWithURLs(StartIndex, i-2-(Length(Value)+1));
                if (Length(Value)=1) and (s[i]<>' ') and (i<>Length(s)) then
                  StartIndex := i
                else
                  StartIndex := i+1;
                StyleNo := GetStyleNo('s', clNone, StyleNo, '', StrToInt(Value));
              end;
              continue;
            end;
          end;
        rsNormal: // normal text, or beginning of fields
          case s[i] of
            '\':
              CodeStarted := not CodeStarted;
            'b', 'u', 'i', 'B', 'U', 'I':
              if CodeStarted then
              begin
                AddStringWithURLs(StartIndex, i-2);
                StartIndex := i+1;
                StyleNo := GetStyleNo(LowerCase(s[i])[1], clNone, StyleNo, '', 0);
                CodeStarted := False;
              end;
            'c', 'C', 'k', 'K':
              if CodeStarted then
              begin
                ColorChar := LowerCase(s[i])[1];
                Value := '';
                ReadState := rsColor;
                CodeStarted := False;
              end;
            'f', 'F':
              if CodeStarted then
              begin
                Value := '';
                ReadState := rsFontNeedsFirstQuote;
                CodeStarted := False;
              end;
            's', 'S':
              if CodeStarted then
              begin
                Value := '';
                ReadState := rsFontSize;
                CodeStarted := False;
              end;
            else
              begin
                CodeStarted := False;
                // emoticons
                EmoticonIndex := EmoticonAtIndex(s, i);
                if EmoticonIndex>=0 then
                begin
                  AddStringWithURLs(StartIndex, i-1);
                  AddEmoticon(EmoticonIndex);
                  StartIndex := i+Length(EmoticonsCodes[EmoticonIndex].Code);
                end;
              end;
          end;
      end;
  AddStringWithURLs(StartIndex, Length(s));
end;
{------------------------------------------------------------------------------}
// OnJump event of rv and rvPreview: opens URL in browser/e-mail client
procedure TForm1.rvJump(Sender: TObject; id: Integer);
var s: TRVUnicodeString;
    RVData: TCustomRVFormattedData;
    ItemNo: Integer;
begin
  TCustomRichView(Sender).GetJumpPointLocation(id, RVData, ItemNo);
  s := RVData.GetItemTextW(ItemNo);
  if not RVIsURL(s) and RVIsEmail(s) then
    s := 'mailto:'+s;
  ShellExecuteW(0, 'open', PRVUnicodeChar(s), nil, nil, SW_SHOW)
end;

end.
